package cn.iocoder.yudao.module.reward.api.reward;

import cn.hutool.core.bean.BeanUtil;
import cn.hutool.core.lang.generator.SnowflakeGenerator;
import cn.iocoder.yudao.framework.common.exception.ServiceException;
import cn.iocoder.yudao.module.member.api.score.MemberUserScoreApi;
import cn.iocoder.yudao.module.member.api.score.dto.MemberUserScoreOperateReqDTO;
import cn.iocoder.yudao.module.member.api.user.MemberUserApi;
import cn.iocoder.yudao.module.member.api.user.dto.UserRespDTO;
import cn.iocoder.yudao.module.member.enums.PlatformTypeEnum;
import cn.iocoder.yudao.module.member.enums.ScoreSourceTypeEnum;
import cn.iocoder.yudao.module.reward.api.reward.dto.RedeemDetailForUserDTO;
import cn.iocoder.yudao.module.reward.api.reward.dto.RedeemRewardReqVO;
import cn.iocoder.yudao.module.reward.api.reward.dto.RedeemRewardRespDTO;
import cn.iocoder.yudao.module.reward.controller.app.redeem.vo.AppRedeemRewardReqVO;
import cn.iocoder.yudao.module.reward.dal.dataobject.redeem.RewardRedeemDO;
import cn.iocoder.yudao.module.reward.dal.dataobject.reward.RewardDO;
import cn.iocoder.yudao.module.reward.dal.mysql.redeem.RewardRedeemMapper;
import cn.iocoder.yudao.module.reward.dal.mysql.reward.RewardMapper;
import cn.iocoder.yudao.module.reward.enums.RewardPickMethedEnum;
import cn.iocoder.yudao.module.reward.enums.RewardRedeemStatusEnum;
import cn.iocoder.yudao.module.reward.enums.RewardStatusEnum;
import cn.iocoder.yudao.module.reward.service.reward.RewardService;
import cn.iocoder.yudao.module.reward.util.RewardGenCodeUtils;
import cn.iocoder.yudao.module.system.api.sms.SmsCodeApi;
import cn.iocoder.yudao.module.system.api.sms.dto.code.SmsCodeSendReqDTO;
import cn.iocoder.yudao.module.system.api.sms.dto.code.SmsCodeUseReqDTO;
import cn.iocoder.yudao.module.system.enums.sms.SmsSceneEnum;
import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.redisson.api.RLock;
import org.redisson.api.RedissonClient;
import org.springframework.aop.framework.AopContext;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.validation.annotation.Validated;

import javax.annotation.Resource;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.TimeUnit;

import static cn.iocoder.yudao.framework.common.exception.util.ServiceExceptionUtil.exception;
import static cn.iocoder.yudao.framework.common.util.servlet.ServletUtils.getClientIP;
import static cn.iocoder.yudao.module.infra.enums.ErrorCodeConstants.GET_LOCK_FAILED;
import static cn.iocoder.yudao.module.member.enums.ErrorCodeConstants.USER_NOT_EXISTS;
import static cn.iocoder.yudao.module.reward.enums.ErrorCodeConstants.*;

@Service
@Validated
@Slf4j
public class RedeemRewardApiImpl implements RedeemRewardApi {

    @Resource
    private RewardMapper rewardMapper;
    @Resource
    private RewardService rewardService;
    @Resource
    private RewardRedeemMapper rewardRedeemMapper;
    @Resource
    private MemberUserApi memberUserApi;
    @Resource
    private RedissonClient redissonClient;
    @Resource
    private MemberUserScoreApi memberUserScoreApi;
    @Resource
    private SnowflakeGenerator snowflakeGenerator;
    @Resource
    private SmsCodeApi smsCodeApi;
    @Resource
    private RewardGenCodeUtils rewardGenCodeUtils;

    @Override
    public Boolean redeemReward(RedeemRewardReqVO redeemRewardReqVO) {
        UserRespDTO memberUser = memberUserApi.getUser(redeemRewardReqVO.getMemberId());
        if (memberUser == null) {
            throw exception(USER_NOT_EXISTS);
        }
        // 验证短信验证码
        if (redeemRewardReqVO.getEntrance() != PlatformTypeEnum.BACKEND.getValue() && redeemRewardReqVO instanceof AppRedeemRewardReqVO) {
            AppRedeemRewardReqVO appRedeemRewardReqVO = (AppRedeemRewardReqVO) redeemRewardReqVO;
            SmsCodeUseReqDTO smsCodeUseReqDTO = new SmsCodeUseReqDTO();
            smsCodeUseReqDTO.setMobile(memberUser.getMobile());
            smsCodeUseReqDTO.setCode(appRedeemRewardReqVO.getCode());
            smsCodeUseReqDTO.setUsedIp(getClientIP());
            smsCodeUseReqDTO.setScene(SmsSceneEnum.MEMBER_REDEEM_REWARD.getScene());
            smsCodeApi.useSmsCode(smsCodeUseReqDTO);
        }
        // 校验收货人信息
        verifyRecipientInfo(redeemRewardReqVO, memberUser);
        // 校验礼品
        RewardDO rewardDO = getAndVerifyRewardDO(redeemRewardReqVO);
        // 校验会员
        verifyMemberUser(redeemRewardReqVO, rewardDO, memberUser.getHoldScore());
        String lockKey = "reward:redeem:lock:" + redeemRewardReqVO.getRewardId();
        RLock lock = redissonClient.getLock(lockKey);
        try {
            if (!lock.tryLock(2, 10, TimeUnit.SECONDS)) {
                throw exception(GET_LOCK_FAILED);
            }
            RedeemRewardApi currentProxy = (RedeemRewardApi) AopContext.currentProxy();
            currentProxy.executeRedeem(redeemRewardReqVO, rewardDO.getPointsRequire(), rewardDO.getId());
        } catch (InterruptedException e) {
            throw exception(GET_LOCK_FAILED);
        } finally {
            lock.unlock();
        }
        return true;
    }

    @Override
    @Transactional(rollbackFor = Exception.class)
    public void executeRedeem(RedeemRewardReqVO redeemRewardReqVO, Integer pointsRequire, Long rewardId) {
        // 添加兑换记录
        RewardRedeemDO redeemRecord = addRedeemRecord(redeemRewardReqVO, pointsRequire);
        // 更新会员积分
        updateMemberScore(redeemRewardReqVO, pointsRequire, redeemRecord.getId(), redeemRecord.getRedemptionNumber());
        // 更新礼品
        redeemReward(rewardId, redeemRewardReqVO.getRewardCount());
    }

    private RewardDO getAndVerifyRewardDO(RedeemRewardReqVO redeemRewardReqVO) {
        // 查询礼品
        RewardDO rewardDO = rewardMapper.selectById(redeemRewardReqVO.getRewardId());
        if (rewardDO == null) {
            throw exception(REWARD_NOT_EXISTS);
        }
        // 礼品未开启
        if (rewardDO.getStatus() != RewardStatusEnum.ENABLED.getValue() || rewardService.validateExpire(rewardDO)) {
            throw exception(REWARD_NOT_ENABLE);
        }
        // 礼品数量不够
        if (rewardDO.getQuantityRemain() < redeemRewardReqVO.getRewardCount()) {
            throw exception(REWARD_COUNT_NOT_ENOUGH);
        }
        // 兑换方式不匹配
        String pickMethod = rewardDO.getPickMethod();
        String[] split = pickMethod.split(",");
        if (!Arrays.asList(split).contains(String.valueOf(redeemRewardReqVO.getRedeemType()))) {
            throw exception(REWARD_PICK_METHOD_NOT_ALLOW);
        }
        return rewardDO;
    }

    private void verifyRecipientInfo(RedeemRewardReqVO redeemRewardReqVO, UserRespDTO memberUser) {
        if (redeemRewardReqVO.getRedeemType() == RewardPickMethedEnum.SELF_PICKUP.getValue()) {
            if (StringUtils.isBlank(redeemRewardReqVO.getRecipientName())) {
                redeemRewardReqVO.setRecipientName(memberUser.getNickname());
            }
            if (StringUtils.isBlank(redeemRewardReqVO.getRecipientPhoneNum())) {
                redeemRewardReqVO.setRecipientPhoneNum(memberUser.getMobile());
            }
        } else if (redeemRewardReqVO.getEntrance() != PlatformTypeEnum.BACKEND.getValue()) {
            if (StringUtils.isAnyBlank(redeemRewardReqVO.getRecipientName(), redeemRewardReqVO.getRecipientPhoneNum(), redeemRewardReqVO.getRecipientAddress())) {
                throw exception(REWARD_REDEEM_RECIPIENT_ERROR);
            }
        }
    }

    private void redeemReward(Long rewardId, Integer rewardCount) {
        // 更新时校验礼品数量是否足够
        RewardDO rewardDO = rewardMapper.selectById(rewardId);
        if (rewardDO.getQuantityRemain() < rewardCount) {
            throw exception(REWARD_COUNT_NOT_ENOUGH);
        }
        RewardDO upReward = new RewardDO();
        upReward.setId(rewardId);
        upReward.setExchangeCount(rewardDO.getExchangeCount() + 1);
        upReward.setQuantityRemain(rewardDO.getQuantityRemain() - rewardCount);
        rewardMapper.updateById(upReward);
    }

    private void updateMemberScore(RedeemRewardReqVO redeemRewardReqVO, Integer pointsRequire, Long redeemId, String redemptionNumber) {
        Map<String, Object> extParam = new HashMap<>();
        extParam.put("redeemId", redeemId);
        extParam.put("redemptionNumber", redemptionNumber);
        memberUserScoreApi.operateScore(MemberUserScoreOperateReqDTO.builder()
                .memberId(redeemRewardReqVO.getMemberId())
                .sourceType(ScoreSourceTypeEnum.EXCHANGE_REWARD)
                .scoreCount(redeemRewardReqVO.getRewardCount() * pointsRequire)
                .releationId(String.valueOf(redeemId))
                .uniqueId(String.valueOf(redeemId))
                .extParam(extParam)
                .build());
    }

    private RewardRedeemDO addRedeemRecord(RedeemRewardReqVO redeemRewardReqVO, Integer pointsRequire) {
        RewardRedeemDO rewardRedeemDO = BeanUtil.copyProperties(redeemRewardReqVO, RewardRedeemDO.class);
        rewardRedeemDO.setId(snowflakeGenerator.next());
        rewardRedeemDO.setRedemptionNumber(rewardGenCodeUtils.generateRedemptionNumber());
        rewardRedeemDO.setStatus(RewardRedeemStatusEnum.REDEEMING.getValue());
        rewardRedeemDO.setScoreCount(redeemRewardReqVO.getRewardCount() * pointsRequire);
        rewardRedeemMapper.insert(rewardRedeemDO);
        return rewardRedeemDO;
    }

    @Override
    public List<RedeemRewardRespDTO> redeemRewards(List<RedeemRewardReqVO> redeemRewardReqVOList) {
        // 批量兑换每次最多十条
        if (redeemRewardReqVOList.size() > 10) {
            throw exception(REWARD_REDEEM_COUNT_NOT_ALLOW);
        }
        Long rewardId = redeemRewardReqVOList.get(0).getRewardId();
        // 查询礼品
        RewardDO rewardDO = rewardMapper.selectById(rewardId);
        if (rewardDO == null) {
            throw exception(REWARD_NOT_EXISTS);
        }
        // 礼品未开启
        if (rewardDO.getStatus() != RewardStatusEnum.ENABLED.getValue() || rewardService.validateExpire(rewardDO)) {
            throw exception(REWARD_NOT_ENABLE);
        }
        int totalCount = 0;
        for (RedeemRewardReqVO redeemRewardReqVO : redeemRewardReqVOList) {
            UserRespDTO memberUser = memberUserApi.getUser(redeemRewardReqVO.getMemberId());
            if (memberUser == null) {
                throw exception(USER_NOT_EXISTS);
            }
            verifyRecipientInfo(redeemRewardReqVO, memberUser);
            // 每个兑换VO校验一遍
            // 兑换方式不匹配，后台不校验
            if (redeemRewardReqVO.getEntrance() != PlatformTypeEnum.BACKEND.getValue()) {
                String pickMethod = rewardDO.getPickMethod();
                String[] split = pickMethod.split(",");
                if (!Arrays.asList(split).contains(String.valueOf(redeemRewardReqVO.getRedeemType()))) {
                    throw exception(REWARD_PICK_METHOD_NOT_ALLOW);
                }
            }
            verifyMemberUser(redeemRewardReqVO, rewardDO, memberUser.getHoldScore());
            // 记录兑换总数
            totalCount += redeemRewardReqVO.getRewardCount();
            // 判断兑换总数是否大于礼物数量
            if (totalCount > rewardDO.getQuantityRemain()) {
                throw exception(REWARD_COUNT_NOT_ENOUGH);
            }
        }
        String lockKey = "reward:redeem:lock:" + rewardId;
        RLock lock = redissonClient.getLock(lockKey);
        try {
            if (!lock.tryLock(2, 10, TimeUnit.SECONDS)) {
                throw exception(GET_LOCK_FAILED);
            }
            RedeemRewardApi currentProxy = (RedeemRewardApi) AopContext.currentProxy();
            currentProxy.batchExecuteRedeem(redeemRewardReqVOList, rewardDO.getPointsRequire(), rewardDO.getId());
        } catch (ServiceException e) {
            log.error("batch redeem rewards exception:{}", e.getMessage());
            throw e;
        } catch (InterruptedException e) {
            throw exception(GET_LOCK_FAILED);
        } finally {
            lock.unlock();
        }
        return null;
    }

    @Override
    @Transactional(rollbackFor = Exception.class)
    public void batchExecuteRedeem(List<RedeemRewardReqVO> redeemRewardReqVOList, Integer pointsRequire, Long rewardId) {
        for (RedeemRewardReqVO redeemRewardReqVO : redeemRewardReqVOList) {
            executeRedeem(redeemRewardReqVO, pointsRequire, rewardId);
        }
    }

    @Override
    public RedeemDetailForUserDTO getRedeemDetailForUser(Long redeemId) {
        RewardRedeemDO rewardRedeemDO = rewardRedeemMapper.selectById(redeemId);
        if (rewardRedeemDO != null) {
            RedeemDetailForUserDTO redeemDetailForUserDTO = new RedeemDetailForUserDTO();
            redeemDetailForUserDTO.setRewardCount(rewardRedeemDO.getRewardCount());
            RewardDO rewardDO = rewardMapper.selectById(rewardRedeemDO.getRewardId());
            redeemDetailForUserDTO.setRewardNameEn(rewardDO.getTitleEn());
            redeemDetailForUserDTO.setRewardNameZh(rewardDO.getTitleZh());
            return redeemDetailForUserDTO;
        }
        return null;
    }

    @Override
    public void sendSmsCode(Long memberId) {
        UserRespDTO memberUserApiUser = memberUserApi.getUser(memberId);
        if (memberUserApiUser == null) {
            throw exception(USER_NOT_EXISTS);
        }
        SmsCodeSendReqDTO smsCodeSendReqDTO = new SmsCodeSendReqDTO();
        smsCodeSendReqDTO.setMobile(memberUserApiUser.getMobile());
        smsCodeSendReqDTO.setAreaCode(memberUserApiUser.getAreaCode());
        smsCodeSendReqDTO.setCreateIp(getClientIP());
        smsCodeSendReqDTO.setScene(SmsSceneEnum.MEMBER_REDEEM_REWARD.getScene());

        smsCodeApi.sendSmsCode(smsCodeSendReqDTO);
    }

    private void verifyMemberUser(RedeemRewardReqVO redeemRewardReqVO, RewardDO rewardDO, Integer holdScore) {
        // 会员积分不够
        if (holdScore < rewardDO.getPointsRequire() * redeemRewardReqVO.getRewardCount()) {
            throw exception(REWARD_SCORE_NOT_ENOUGH);
        }
        // 校验兑换次数
        LambdaQueryWrapper<RewardRedeemDO> rewardRedeemDOWrapper = new LambdaQueryWrapper<>();
        rewardRedeemDOWrapper.eq(RewardRedeemDO::getRewardId, redeemRewardReqVO.getRewardId())
                .eq(RewardRedeemDO::getMemberId, redeemRewardReqVO.getMemberId());
        Long count = rewardRedeemMapper.selectCount(rewardRedeemDOWrapper);
        if (count >= rewardDO.getAllowCount()) {
            throw exception(REWARD_REDEEM_ALLOW_COUNT_ERROR);
        }
    }
}
